{-# LANGUAGE DeriveDataTypeable #-}

module KoL.UtilTypes where

import Prelude hiding (read, catch)

import qualified Data.Map
import Data.IORef
import Data.Time
import Control.Concurrent
import Control.Monad

import System.IO
import Network.URI
import qualified Database.SQLite3
import qualified Network.HTTP
import qualified Data.ByteString
import Control.Exception
import Data.Typeable

import qualified Scripting.LuaModded

type DiscerningStateIdentifier = (String, Integer, Integer, String) -- name, ascension, rollover, sessid

type StateType = (Data.Map.Map String String, Data.Map.Map String String, Data.Map.Map String String, Data.Map.Map String String, Data.Map.Map String String)

-- This entire thing should be locked up into DiscerningStateIdentifier. log/state actions don't really belong.
data SessionDataType = SessionDataType {
	jsonStatusPageMVarRef :: IORef (MVar (Maybe String)),
	lastRetrieve :: IORef UTCTime,
	itemData :: IORef (Maybe (Int -> Maybe [(String, String)], String -> Maybe [(String, String)])),
	doDbLogAction :: RefType -> (Database.SQLite3.Database -> IO ()) -> IO (),
	doStateAction :: RefType -> (Database.SQLite3.Database -> IO ()) -> IO (),
	stateData :: IORef (Maybe (DiscerningStateIdentifier, StateType)),
	luaInstances :: IORef (Data.Map.Map (String, String) (MVar Scripting.LuaModded.LuaState))
}

type ConnChanActionType = MVar (Either SomeException (URI, Data.ByteString.ByteString, [(String, String)], Network.HTTP.ResponseCode, Network.HTTP.Response Data.ByteString.ByteString))
type ConnChanType = Chan (URI, Network.HTTP.Request Data.ByteString.ByteString, ConnChanActionType, RefType)

data ServerSessionType = ServerSessionType {
	wheneverConnection :: ConnChanType,
	sequenceConnection :: ConnChanType,
	sessConnData :: SessionDataType
}

data ConnectionType = ConnectionType {
	cookie :: Maybe String,
	useragent :: String,
	hostUri :: URI,
	getconn :: ConnChanType
}

data LogRefStuff = LogRefStuff {
	logchan_ :: Chan (IO ()),
	solid_logchan_ :: Chan (IO ()),
	logcontext_ :: MVar String
}

data ProcessingRefStuff = ProcessingRefStuff {
	processPage_ :: RefType -> URI -> Maybe [(String, String)] -> IO (IO (Either (Data.ByteString.ByteString, URI) (Data.ByteString.ByteString, URI))),
	rawRetrievePage_ :: RefType -> URI -> Maybe [(String, String)] -> IO (IO (Data.ByteString.ByteString, URI, [(String, String)])),
	nochangeRawRetrievePageFunc_ :: RefType -> URI -> Maybe [(String, String)] -> Bool -> IO (IO (Data.ByteString.ByteString, URI, [(String, String)])),
	getstatusfunc_ :: RefType -> IO (IO String)
}

data OtherRefStuff = OtherRefStuff {
	connection_ :: ConnectionType,
	sessionData_ :: SessionDataType,
	logindents_ :: IORef Integer
}

data RefType = RefType {
	logstuff_ :: LogRefStuff,
	processingstuff_ :: ProcessingRefStuff,
	otherstuff_ :: OtherRefStuff,
	stateValid_ :: Bool
}

logchan ref = logchan_ $ logstuff_ $ ref
logcontext ref = logcontext_ $ logstuff_ $ ref

processPage ref = processPage_ $ processingstuff_ $ ref
rawRetrievePage ref = rawRetrievePage_ $ processingstuff_ $ ref
nochangeRawRetrievePageFunc ref = nochangeRawRetrievePageFunc_ $ processingstuff_ $ ref
readstatus ref = join $ getstatusfunc ref
getstatusfunc ref = (getstatusfunc_ $ processingstuff_ $ ref) ref
refreshstatus ref = void $ join $ getstatusfunc ref

connection ref = connection_ $ otherstuff_ $ ref
state ref = case stateValid_ ref of
	True -> stateData $ sessionData $ ref
	_ -> throw $ InternalError $ "Invalid state while trying to read"
sessionData ref = sessionData_ $ otherstuff_ $ ref
logindents ref = logindents_ $ otherstuff_ $ ref

data LogItem = LogItem {
	time :: ZonedTime,
	apiStatusBefore :: String,
	apiStatusAfter :: String,
	stateBefore :: (Maybe StateType),
	stateAfter :: (Maybe StateType),
	sessionId :: String,
	requestedUri :: URI,
	parameters :: Maybe [(String, String)],
	retrievedUri :: URI,
	pageText :: String
}

data ConnectionException = UrlMismatchException String URI | NotLoggedInException | ApiPageException String | HttpRequestException URI SomeException
	deriving (Typeable)

instance Exception ConnectionException

instance Show ConnectionException where
	show (UrlMismatchException urlstr goturi) = "Error loading URL: " ++ urlstr ++ ", received: " ++ (show goturi)
	show (NotLoggedInException) = "Not logged in"
	show (ApiPageException pagestr) = "Error loading API, page returned: " ++ pagestr
	show (HttpRequestException uri err) = "Network connection error while loading " ++ uriPath uri ++ " (exception: " ++ show err ++ ")"

data InternalError = InternalError String
	deriving (Typeable, Show)

instance Exception InternalError
