{-# LANGUAGE BangPatterns #-}

-- module LogParser (showLogs, compressLogs, parse_log, do_log_parsing, parse_line_data, parse_page_title, parse_line) where
module LogParser (showLogs, compressLogs) where

-- TODO: drill down to whether a page is useful ASAP, instead of having long lists around and parsing them
-- TODO: store parts of status separately??? means text gets non-identical

import Prelude hiding (read, catch)

import Data.Maybe
import Data.List
import KoL.Util
import System.IO
import Network.URI
import Text.Printf
import System.Directory
import Control.Applicative
import Text.JSON
import Control.Monad
import Control.Concurrent
import Control.Exception
import Data.IORef

import qualified Codec.Compression.BZip
import qualified Data.ByteString.Lazy.Char8
import qualified Data.ByteString.Lazy.Internal
import qualified Database.SQLite3




data LogPageload = LogPageload {
	logidx :: !Integer,
	uri :: !URI,
	effuri :: !URI,
-- 	title :: Maybe String,
	turn_before :: !Integer,
	turn_after :: !Integer,
	ascensionday :: !Integer,
	zonename :: !String,
	zonecontainer :: !String,
-- 	allparams :: [(String, String)],
	stats_before :: !(Integer, Integer, Integer),
	stats_after :: !(Integer, Integer, Integer),
	trail_before :: !Bool,
	trail_after :: !Bool
} deriving (Show)

parse_page_title pagetext =
	case (pagetitle, monstername, advrestitle) of
		(_, [[x]], _) -> Just $ x
		(_, _, [[x]]) -> Just $ x
		([["Results:"]], _, _) -> Nothing
		([["Adventure Results:"]], _, _) -> Nothing
		([[x]], _, _) -> Just $ x
		_ -> Nothing
	where
		pagetitle = matchGroups "<tr><td style=\"color: white;\" align=center bgcolor=blue><b>([^<]+)</b></td></tr>" pagetext
		monstername = matchGroups ">You're fighting <span id='monname'>([^<]*)</span><" pagetext
		advrestitle = matchGroups "<td style=\"color: white;\" align=center bgcolor=blue><b>[a-zA-Z]* Results:</b></td></tr><tr><td style=\"padding: 5px; border: 1px solid blue;\"><center><table><tr><td><center><b>([^<]*)</b>" pagetext

format_number_range from to
	| (from == to) = "#"
	| (from == to - 1) = show from
	| (from < to) = (show $ from) ++ "-" ++ (show $ to - 1)

getstatus jscombinedstr = jso where
	Ok jscombobj = decodeStrict jscombinedstr :: Result (JSObject JSValue)
	Ok jso = valFromObj "status" jscombobj :: Result (JSObject JSValue)

get_value name jsobj = x where
	Just (JSString jsstr) = lookup name $ fromJSObject jsobj
	!x = fromJSString jsstr

get_value_int name jsobj = case valFromObj name jsobj of
	Ok (JSString s) -> Just jss where
		Just !jss = read_as (fromJSString s)
	Ok (JSRational _ r) -> Just rounded where
		!rounded = round r
	Ok JSNull -> Just 0 -- HACK for when rollover is broken and not set yet
	_ -> Nothing

isontrail jsobj = case valFromObj "effects" jsobj :: Result (JSObject JSValue) of
	Ok effectsobj -> isJust $ lookup "91635be2834f8a07c8ff9e3b47d2e43a" $ fromJSObject effectsobj
	_ -> False

build_thing lidx requri returi jsbefore jsafter __raw_row_cols = do
	if get_value_int "freedralph" jsbefore == Just 1
		then return Nothing
		else (do
			let Ok lastadvobjafter = valFromObj "lastadv" jsafter
			let getstats jsobj = (fromJust $ get_value_int "rawmuscle" jsobj, fromJust $ get_value_int "rawmysticality" jsobj, fromJust $ get_value_int "rawmoxie" jsobj)
			let !thing = LogPageload {
				logidx = fromJust $ read_as $ lidx,
				uri = requri,
				effuri = returi,
				turn_before = (fromJust $ get_value_int "turnsthisrun" jsbefore) + 1,
				turn_after = (fromJust $ get_value_int "turnsthisrun" jsafter) + 1,
				ascensionday = case get_value_int "daysthisrun" jsbefore of
					Just d -> d
					_ -> fromJust $ get_value_int "rollover" jsbefore,
				zonename = get_value "name" lastadvobjafter,
				zonecontainer = get_value "container" lastadvobjafter,
				stats_before = getstats jsbefore,
				stats_after = getstats jsafter,
				trail_before = isontrail jsbefore,
				trail_after = isontrail jsafter
			}
			let !__l = length $ show $ thing
			return $ Just thing) `catch` (\e -> do
					putStrLn $ "Error parsing row: " ++ show lidx ++ ": " ++ show (e :: SomeException)
					putStrLn $ "  contents: " ++ show __raw_row_cols
					return Nothing)

go_build_thing __raw_row_cols@[lidx, requestedurlstr, retrievedurlstr, statusbeforestr, statusafterstr] = do
	let Just requri = parseURIReference requestedurlstr
	let Just returi = parseURIReference retrievedurlstr
	let !jsbefore = getstatus statusbeforestr
	let !jsafter = getstatus statusafterstr
	build_thing lidx requri returi jsbefore jsafter __raw_row_cols

getresults s acc = do
	sr <- Database.SQLite3.step s
	case sr of
		Database.SQLite3.Row -> do
			[Database.SQLite3.SQLInteger lidx, Database.SQLite3.SQLText requestedurlstr, Database.SQLite3.SQLText retrievedurlstr, Database.SQLite3.SQLText statusbeforestr, Database.SQLite3.SQLText statusafterstr] <- Database.SQLite3.columns s
			newthing <- go_build_thing [show lidx, requestedurlstr, retrievedurlstr, statusbeforestr, statusafterstr]
			case newthing of
				Just x -> getresults s (x:acc)
				_ -> getresults s acc
		Database.SQLite3.Done -> return acc

scan_through_database filename = do
	log_db <- Database.SQLite3.open filename
	rsgood <- do
		s <- Database.SQLite3.prepare log_db "SELECT idx, requestedurl, retrievedurl, statusbefore, statusafter FROM pageloads;"
		r <- reverse <$> getresults s []
		Database.SQLite3.finalize s
		return r

	putStrLn $ filename ++ " length rsgood: " ++ (show $ length $ rsgood)

	let group_by_day x y = ascensionday x == ascensionday y
	let group_by_container x y = zonecontainer x == zonecontainer y
	let group_by_zone x y = zonename x == zonename y
	let group_by_turn x y = turn_before x == turn_before y
	let format_turn_range format xs =
			"<span class=\"turnrange inside" ++ extraclass ++ "\">" ++ (printf format tstr) ++ "</span>"
		where
			tstr = format_number_range from to
			from = turn_before (head xs)
			to = turn_after (last xs)
			extraclass = if format == "(%s)" then " pageload" else ""
	let desc_statgain xs = printf "(%+d mus, %+d mys, %+d mox)" (samus - sbmus) (samys - sbmys) (samox - sbmox) :: String
		where
			(sbmus, sbmys, sbmox) = stats_before (head xs)
			(samus, samys, samox) = stats_after (last xs)
	let do_line (a, b, c) = return $ printf "%s %s" (format_turn_range c b) a :: IO String
	let do_turn turns = do
		let get_title rem gottype = case rem of
			[] -> return $ "??? " ++ (fromJust gottype) ++ " ???"
			(x:xs) -> do
				[[Just pagetext]] <- do_db_query log_db "SELECT pagetext FROM pageloads WHERE idx = ?;" [Just $ show $ logidx $ x]
				let xtype = (uriPath $ effuri $ x)
				case (parse_page_title pagetext, turn_before x == turn_after x, xtype `elem` ["/fight.php", "/choice.php", "/adventure.php"], gottype == Nothing || gottype == Just xtype) of
					(Just z, False, _, _) -> return z
					(Just z, _, True, _) -> return z
					(_, _, True, True) -> get_title xs (Just xtype)
					_ -> return $ "??? " ++ (fromMaybe "?" gottype) ++ " <- " ++ xtype ++ " ???"

		if (turn_before $ head $ turns) == (turn_after $ last $ turns)
			then return []
			else do
				!y <- get_title (reverse turns) Nothing
				!t <- do_line (y, turns, "(%s)")
				return [t ++ "<br>\n"]

	zdivnamecounter <- newIORef (1 :: Integer)
	cdivnamecounter <- newIORef (1 :: Integer)
	ddivnamecounter <- newIORef (1 :: Integer)

	let do_level shortname do_sub namecounter group_method headliner xs = do
		let sn = shortname
		xid <- readIORef namecounter
		writeIORef namecounter (xid + 1)
		let subxs = groupBy group_method xs
		let desc = printf "%s %s" (format_turn_range "[%s]" xs) (headliner xs) :: String
		subret <- mapM do_sub subxs
		subret <- return $ concat subret
		case subret of
			[] -> do
				ret <- return $ []
				ret <- return $ (++) ret $ printf "<div id=\"%s%d\" class=\"%s\">\n" sn xid sn
				ret <- return $ (++) ret $ printf "%s %s<br>\n" desc (desc_statgain xs)
				ret <- return $ (++) ret $ "</div>\n"
				return [ret]
			[sr] -> return [sr]
			_ -> do
				ret <- return $ []
				ret <- return $ (++) ret $ printf "<div id=\"%s%d\" class=\"%s\">\n" sn xid sn
				ret <- return $ (++) ret $ printf "%s %s <a href=\"#\" onclick=\"return toggle(this, %sd%d);\" class=\"expander\">[+]</a><br>\n" desc (desc_statgain xs) sn xid
				ret <- return $ (++) ret $ printf "  <div id=\"%sd%d\" class=\"details\">\n" sn xid
				ret <- return $ ret ++ (concat subret)
				ret <- return $ (++) ret $ "  </div>\n"
				ret <- return $ (++) ret $ "</div>\n"
				return [ret]

	let do_zone = do_level "z" do_turn zdivnamecounter group_by_turn (\xs -> zonename $ head $ xs)
	let do_container = do_level "c" do_zone cdivnamecounter group_by_zone (\xs -> zonecontainer $ head $ xs)
	let do_day = do_level "d" do_container ddivnamecounter group_by_container (\xs -> printf "Day %d" $ ascensionday $ head $ xs :: String)

	all_lines <- mapM do_day $ groupBy group_by_day rsgood

	let header_html = [
		"<html>",
		"<head>",
		"  <style>",
		"    body { font-family: monospace; }",
		"    .expander { color: green; }",
		"    .turnrange { color: blue; }",
		"    .turnrange.pageload { color: darkcyan; }",
		"    .inside { min-width: 12em; }",
		"    .details { margin-left: 20px; color: dimgray; display: none; }",
		"  </style>",
		"  <script type=\"text/javascript\">",
		"    function toggle(t, id) {",
		"      if (id.style.display == 'block') {",
		"        t.innerHTML = '[+]'",
		"        id.style.display = 'none'",
		"        return false",
		"      } else {",
		"        t.innerHTML = '[-]'",
		"        id.style.display = 'block'",
		"        return false",
		"      }",
		"    }",
		"  </script>",
		"</head>",
		"<body>"]

	let footer_html = [
		"</body>",
		"</html>"]

	Database.SQLite3.close log_db

	return $ header_html ++ (concat all_lines) ++ footer_html



yielding_bs_eval Data.ByteString.Lazy.Internal.Empty = return Data.ByteString.Lazy.Internal.Empty
yielding_bs_eval (Data.ByteString.Lazy.Internal.Chunk x y) = do
	threadDelay 100000 -- delay to allow other things to run
	yy <- yielding_bs_eval y
	return $ Data.ByteString.Lazy.Internal.Chunk x yy





showLogs fn pwd = do
	case fn of
		Just basefilename -> do
			bz2path <- getDirectoryPath "sqlite3 log" (basefilename ++ ".ascension-log.sqlite3.bz2")
			bz2exists <- doesFileExist bz2path
			lines <- case bz2exists of
				False -> do
					path <- getDirectoryPath "sqlite3 log" (basefilename ++ ".ascension-log.sqlite3")
					putStrLn $ "processing " ++ path
					withBinaryFile "temp-logparse-file.sqlite3" WriteMode $ \handle_out -> do
						withBinaryFile path ReadMode $ \handle_in -> do
							Data.ByteString.Lazy.Char8.hGetContents handle_in >>= Data.ByteString.Lazy.Char8.hPut handle_out
					putStrLn $ "  parsing log"
					scan_through_database "temp-logparse-file.sqlite3"
				True -> do
					putStrLn $ "processing compressed " ++ bz2path
					withBinaryFile "temp-logparse-file.sqlite3" WriteMode $ \handle_out -> do
						withBinaryFile bz2path ReadMode $ \handle_in -> do
							Codec.Compression.BZip.decompress <$> Data.ByteString.Lazy.Char8.hGetContents handle_in >>= Data.ByteString.Lazy.Char8.hPut handle_out
					putStrLn $ "  parsing log"
					scan_through_database "temp-logparse-file.sqlite3"
			let htmltext = intercalate "\n" lines
			writeFile ("parsed-log-" ++ basefilename ++ ".html") htmltext
			putStrLn $ "log parsing done!"
			return htmltext
		_ -> do
			basedir <- getBaseDirectory "sqlite3 log"
			filenames <- getDirectoryContents basedir
			let describe x =
				case matchGroups "^(.+)-([0-9]+)(.ascension-log.sqlite3|.ascension-log.sqlite3.bz2)$" x of
					[[charname, ascnum, ".ascension-log.sqlite3"]] -> Just (charname, fromJust $ read_as ascnum :: Integer, False)
					[[charname, ascnum, ".ascension-log.sqlite3.bz2"]] -> Just (charname, fromJust $ read_as ascnum :: Integer, True)
					_ -> Nothing
			let ascs = sortBy (\(achar, aasc, _) (bchar, basc, _) -> compare (achar, (-aasc)) (bchar, (-basc))) $ mapMaybe describe filenames
			let desc_asc_link (charname, ascnum, compressed) = case compressed of
					True -> "<small>" ++ link ++ "</small>"
					False -> link
				where
					link = "<a href=\"custom-logs?which=" ++ (charname ++ "-" ++ show ascnum) ++ "&pwd=" ++ pwd ++ "\">" ++ (show ascnum) ++ "</a>"
			let group_to_row x = case x of
				[] -> ""
				(achar, _, _):_ -> "<tr><td>" ++ achar ++ "</td><td>" ++ (intercalate ", " (map desc_asc_link x)) ++ "</td></tr>\n"
			let grouped = groupBy (\(achar, _, _) (bchar, _, _) -> achar == bchar) ascs
			let rowtexts = map group_to_row grouped
			let pt = "<html><body>Click on a log to parse it. Parsing a log takes several minutes, and the format and features are still being developed.<table>\n" ++ (concat rowtexts) ++ "</table></body></html>"
			return pt

compressFile path = do
	let bz2path = path ++ ".bz2"
	bz2exists <- doesFileExist bz2path
	if bz2exists
		then putStrLn $ "Error: " ++ (show bz2path) ++ " already exists."
		else do
			putStrLn $ "compressing " ++ path
			contents <- Data.ByteString.Lazy.Char8.readFile path
			compressed_contents <- yielding_bs_eval $ Codec.Compression.BZip.compress contents
			h <- openFile bz2path WriteMode
			Data.ByteString.Lazy.Char8.hPut h compressed_contents
			hClose h
			putStrLn $ "compressed " ++ path

			test_uncompressed <- Data.ByteString.Lazy.Char8.readFile path
			test_compressed <- Data.ByteString.Lazy.Char8.readFile bz2path
			decompressed <- yielding_bs_eval test_compressed >>= return . Codec.Compression.BZip.decompress
			if test_uncompressed == decompressed
				then do
					putStrLn $ "compression ok, removing file " ++ path
					removeFile path
				else putStrLn $ "Error: compressed file is different from the original."

compressLogs charname asc = do
	let check_file x = do
		path <- getDirectoryPath "sqlite3 log" (charname ++ "-" ++ (show x) ++ ".ascension-log.sqlite3")
		exists <- doesFileExist path
-- 		putStrLn $ "check " ++ show (path, exists)
		when exists $ compressFile path
		
-- 	putStrLn $ "compressing old logs... " ++ charname ++ ": " ++ (show [1..(asc-1)])
	mapM_ check_file [1..(asc-1)]
