{-# LANGUAGE ForeignFunctionInterface #-}

-- TODO: level should be an enum, not a string

module Lua (runProcessScript, runPrinterScript, runChatScript, runAutomateScript, runInterceptScript) where

import Prelude hiding (read, catch)

import KoL.Connection
import KoL.Info
import qualified KoL.Api
import KoL.Util
import KoL.UtilTypes
import State
import Text.JSON
import Control.Exception
import Control.Applicative
import Logging
import HardcodedGameStuff

import Data.IORef
import Data.Maybe
import Network.URI
import Network.CGI
import Control.Monad
import Control.Concurrent

import qualified Data.ByteString.Char8
import qualified Scripting.LuaModded as Lua

import qualified Data.Map

import Text.Printf

import Data.List

doLUA_DEBUG _ = return ()
-- doLUA_DEBUG x = putStrLn $ "LUA DEBUG: " ++ x

set_state ref stateset var value = do
	doLUA_DEBUG $ "lua: writing state " ++ stateset ++ " ~ " ++ var ++ " = " ++ value
	canread <- canReadState ref
	unless canread $ fail $ "Error: Trying to set state \"" ++ var ++ "\" before state is available."
	let do_set ss v = case value of
		"" -> unsetState ref ss v
		val -> setState ref ss v val
	case (stateset, stateset `elem` ["ascension", "day", "session"]) of -- "request"
		("fight", _) -> do_set "day" ("fight." ++ var)
		(_, True) -> do_set stateset var
		_ -> fail $ "cannot write to stateset " ++ (show $ stateset)

get_state ref stateset var = do
	canread <- canReadState ref
	unless canread $ fail $ "Error: Trying to get state \"" ++ var ++ "\" before state is available."
	let do_get ss v = do
		val <- getState ref ss v
		return $ fromMaybe "" val
	val <- case (stateset, stateset `elem` ["character", "ascension", "day", "session"]) of -- "request"
		("fight", _) -> do_get "day" ("fight." ++ var)
		(_, True) -> do_get stateset var
		_ -> fail $ "cannot read stateset " ++ (show $ stateset)
	doLUA_DEBUG $ "lua: getting state " ++ stateset ++ " ~ " ++ var ++ " <- " ++ (show val)
	return val

set_state_whenever ref var value = do
-- 	putStrLn $ "lua: set_state..."  ++ (show $ (var, value))
	charname <- get_ref_playername ref
-- 	putStrLn $ "lua: setting state " ++ var ++ " => " ++ (show value) ++ " for " ++ charname
	chatmap <- readMapFromFile ("chat-" ++ charname ++ ".state")
	chatmap <- return $ Data.Map.insert var value chatmap
	writeMapToFile ("chat-" ++ charname ++ ".state") chatmap
	return ()

get_state_whenever ref var = do
-- 	putStrLn $ "lua: get_state..." ++ (show $ (var))
	charname <- get_ref_playername ref
-- 	putStrLn $ "lua: getting " ++ (show var) ++ " for " ++ charname
	chatmap <- readMapFromFile ("chat-" ++ charname ++ ".state")
	let val = Data.Map.lookup var chatmap
	case val of
		Nothing -> return $ ""
		Just x -> return $ x

get_player_id ref name = do
	doLUA_DEBUG $ "getting ID for " ++ name
	pid <- getPlayerId name ref
	doLUA_DEBUG $ "  => " ++ (show pid)
	case pid of
		Just x -> return (show x)
		Nothing -> return "-1"

-- TODO: Check if this is really OK. It's not in valhalla!
get_ref_playername ref = KoL.Api.charName <$> KoL.Api.getApiInfo ref

-- parse_table_string :: String -> IO String
parse_table_string str = do
	case read_as str :: Maybe [(String, String)] of
		Just xs -> do
			let luaxs = map (\(idx, (x,y)) -> printf "[%d] = { key = %s, value = %s }" (idx :: Integer) (show x) (show y)) (zip [1..] xs)
			return $ printf "return { %s }" $ intercalate ", " luaxs
		_ -> do
			putStrLn $ "parse_table_string error: " ++ (show str)
			return ""

get_item_funcs_internal ref = do
	let ioref = itemData $ sessionData ref
	v <- readIORef ioref
	case v of
		Just i -> return i
		Nothing -> do
			itemstr <- readFile "cache/data/items"
			let Just items = read_as itemstr :: Maybe [[(String, String)]]
			let idmap = Data.Map.fromList (map (\x -> (fromJust $ read_as $ fromJust $ lookup "id" x :: Int, x)) items)
			let namemap = Data.Map.fromList (map (\x -> (fromJust $ lookup "name" x, x)) items)
			let fs = (\x -> Data.Map.lookup x idmap, \x -> Data.Map.lookup x namemap)
			writeIORef ioref (Just fs)
			return fs

-- TODO: rewrite to return Lua stuff
get_recipes ref = lua_log_line ref "get_recipes" $ readFile "cache/data/recipes"

submit_page ref method inputuristr params = lua_log_line ref "submit_page" $ do
	doLUA_DEBUG $ "> submit_page(" ++ method ++ ", " ++ inputuristr ++ ", " ++ (show params) ++ ")"
	x <- lua_log_line ref "submit_page results" $ case parseURIReference inputuristr of	
		Just inputuri -> do
			case (stripPrefix "/" (uriPath inputuri), uriQuery inputuri) of
				(Just _, "") -> do
					case method of
						"GET" -> lua_log_line ref "submit_page GET" $  do
							case params of
								Nothing -> getUriPage (uriPath inputuri) ref
								Just p -> do
									let testurienc = (uriPath inputuri) ++ "?" ++ (formEncode p)
									let testuri = escapeURIString (\x -> x `notElem` "[]") testurienc
									case parseURIReference testuri of
										Just uri -> getUriPage (show uri) ref
										_ -> fail $ "submit_page error: uri " ++ (show testuri) ++ " not recognized."
						"POST" -> lua_log_line ref "submit_page POST" $ do
							case params of
								Just p -> postUriPage (uriPath inputuri) p ref
								_ -> fail $ "submit_page error: parameters " ++ (show params) ++ " not recognized."
						_ -> fail $ "submit_page error: unknown method " ++ (show method) ++ ", should be GET or POST"
				_ -> fail $ "submit_page error: unknown url " ++ (show inputuri)
		_ -> fail $ "submit_page error: unknown url " ++ (show inputuristr)
	doLUA_DEBUG $ "< submit_page(" ++ method ++ ", " ++ inputuristr ++ ", " ++ (show params) ++ ")"
	-- TODO: must fix this, sometimes it's super super super slow to return? when? why? does this still happen?
	return x


async_submit_page ref method inputuristr params = lua_log_line ref "async_submit_page" $ do
	doLUA_DEBUG $ "> async_submit_page(" ++ method ++ ", " ++ inputuristr ++ ", " ++ (show params) ++ ")"
	x <- lua_log_line ref "async_submit_page results" $ case parseURIReference inputuristr of	
		Just inputuri -> do
			case (stripPrefix "/" (uriPath inputuri), uriQuery inputuri) of
				(Just _, "") -> do
					(final_url, final_params) <- case method of
						"GET" -> lua_log_line ref "async_submit_page GET" $  do
							case params of
								Nothing -> return (uriPath inputuri, Nothing)
								Just p -> do
									let testurienc = (uriPath inputuri) ++ "?" ++ (formEncode p)
									let testuri = escapeURIString (\x -> x `notElem` "[]") testurienc
									case parseURIReference testuri of
										Just uri -> return (show uri, Nothing)
										_ -> fail $ "async_submit_page error: uri " ++ (show testuri) ++ " not recognized."
						"POST" -> lua_log_line ref "async_submit_page POST" $ do
							case params of
								Just p -> return (uriPath inputuri, Just p)
								_ -> fail $ "async_submit_page error: parameters " ++ (show params) ++ " not recognized."
						_ -> fail $ "async_submit_page error: unknown method " ++ (show method) ++ ", should be GET or POST"
					xf <- (processPage ref) ref (mkuri final_url) final_params
					return $ do
						x <- xf
						return $ either id id x
				_ -> fail $ "async_submit_page error: unknown url " ++ (show inputuri)
		_ -> fail $ "async_submit_page error: unknown url " ++ (show inputuristr)
	doLUA_DEBUG $ "< async_submit_page(" ++ method ++ ", " ++ inputuristr ++ ", " ++ (show params) ++ ")"
	return x


make_href _ref inputuristr params = do
	doLUA_DEBUG $ "make_href(" ++ inputuristr ++ ", " ++ (show params) ++ ")"
	case parseURIReference inputuristr of	
		Just inputuri -> do
			case (stripPrefix "/" (uriPath inputuri), uriQuery inputuri) of
				(Just _, "") -> do
					case params of
						Nothing -> return (show inputuri)
						Just p -> do
							let testurienc = (uriPath inputuri) ++ "?" ++ (formEncode p)
							let testuri = escapeURIString (\x -> x `notElem` "[]") testurienc
							case parseURIReference testuri of
								Just uri -> return (show uri)
								_ -> fail $ "make_href error: uri " ++ (show testuri) ++ " not recognized."
				_ -> fail $ "make_href error: unknown url " ++ (show inputuri)
		_ -> fail $ "make_href error: unknown url " ++ (show inputuristr)

push_stringlist l lst = do
	Lua.newtable l
	mapM_ (\(x, y) -> do
		Lua.pushinteger l x
		Lua.pushstring l y
		Lua.settable l (-3)) (zip [1..] lst)

push_integertable l tbl = do
	Lua.newtable l
	mapM_ (\(x, y) -> do
		Lua.pushinteger l (fromIntegral (x :: Integer))
		Lua.pushinteger l (fromIntegral (y :: Integer))
		Lua.settable l (-3)) tbl

push_stringtable l tbl = do
	Lua.newtable l
	mapM_ (\(x, y) -> do
		Lua.pushstring l x
		Lua.pushstring l y
		Lua.settable l (-3)) tbl

push_jsvalue l jsval = do
-- 	putStrLn $ "push_jsvalue: " ++ (show jsval)
	case jsval of
		JSNull -> Lua.pushnil l
		JSBool b -> Lua.pushboolean l b
		JSRational _ r -> Lua.pushnumber l (fromRational r)
		JSString jss -> Lua.pushstring l (fromJSString jss)
		JSArray jsarr -> do
			Lua.newtable l
			mapM_ (\(x, y) -> do
				Lua.pushinteger l x
				push_jsvalue l y
				Lua.settable l (-3)) (zip [1..] jsarr)
		JSObject jsobj -> do
			let m = fromJSObject jsobj :: [(String, JSValue)]
			Lua.newtable l
			mapM_ (\(x, y) -> do
				Lua.pushstring l x
				push_jsvalue l y
				Lua.settable l (-3)) m

push_function l f identifier = do
	Lua.pushhsfunction_raw l identifier (\l -> (do
		x <- f l
		return x) `catch` (\e -> do
			putStrLn $ "lua.hs error in " ++ identifier ++ ": " ++ (show e)
			Lua.pushstring l (show (e :: SomeException))
			return (-1)))

setup_lua_instance level filename ref = do
	(l, c) <- log_time_interval ref ("setup lua: " ++ filename) $ do
		code <- readFile filename
		l <- do
			l <- Lua.newstate
			Lua.openlibs l
			return l

		do
			let tostringtable l idx = do
				let get_more = do
					n <- Lua.next l idx
					if n
						then do
							t1 <- Lua.ltype l (-2)
							t2 <- Lua.ltype l (-1)
							(k, v) <- case (t1, t2) of
								(Lua.TSTRING, Lua.TSTRING) -> do
									k <- Lua.tostring l (-2)
									v <- Data.ByteString.Char8.unpack <$> Lua.tobytestring l (-1)
									return (k, v)
								_ -> fail $ "Table keys/values are not strings: " ++ (show $ (t1, t2))
							Lua.pop l 1
							more <- get_more
							return $ [(k, v)] ++ more
						else return []
				Lua.pushnil l
				get_more

			let toparamstable l idx = do
				let get_more n = do
					Lua.pushinteger l n
					Lua.gettable l idx
					t <- Lua.ltype l (-1)
					case t of
						Lua.TTABLE -> do
							topidx <- Lua.gettop l
							tbl <- tostringtable l topidx
							(k, v) <- case (lookup "key" tbl, lookup "value" tbl) of
								(Just k, Just v) -> return (k, v)
								_ -> fail "key or value missing from param table"
							Lua.pop l 1
							more <- get_more (n + 1)
							return $ [(k, v)] ++ more
						Lua.TNIL -> return []
						_ -> fail "Wrong types for params table"
				get_more 1

			let submit_page_func l = do
				lua_log_line ref "> submit_page_func" (return ())
				top <- Lua.gettop l
				if top < 2
					then fail "Not enough parameters to submit_page"
					else do
						m_one <- Lua.peek l 1
						m_two <- Lua.peek l 2
						case (m_one, m_two) of
							(Just one, Just two) -> do
								params <- if top >= 3
									then do
										is3nil <- Lua.isnil l 3
										is3tbl <- Lua.istable l 3
										case (is3nil, is3tbl) of
											(True, _) -> return Nothing
											(False, True) -> do
												tbl <- toparamstable l 3
												return $ Just tbl
											(False, False) -> fail "submit_page parameter 3 is not a table"
									else return Nothing
								(pt, puri) <- submit_page ref one two params
								Lua.pushbytestring l pt
								Lua.pushstring l (show puri)
								lua_log_line ref ("< submit_page_func " ++ (show puri)) (return ())
								return 2
							_ -> fail "Wrong parameters to submit_page"

			let make_href_func l = do
				top <- Lua.gettop l
				if top < 1
					then fail "Not enough parameters to make_href"
					else do
						m_one <- Lua.peek l 1
						case m_one of
							Just one -> do
								params <- if top >= 2
									then do
										is2nil <- Lua.isnil l 2
										is2tbl <- Lua.istable l 2
										case (is2nil, is2tbl) of
											(True, _) -> return Nothing
											(False, True) -> do
												tbl <- toparamstable l 2
												return $ Just tbl
											(False, False) -> fail "make_href parameter 2 is not a table"
									else return Nothing
								href <- make_href ref one params
								Lua.pushstring l href
								return 1
							_ -> fail "Wrong parameters to make_href"

			let register_function name f = do
				push_function l f name
				Lua.setglobal l name

			register_function "make_href" make_href_func

			let async_submit_page_func l = do
				lua_log_line ref "> async_submit_page_func" (return ())
				top <- Lua.gettop l
				if top < 2
					then fail "Not enough parameters to async_submit_page"
					else do
						m_one <- Lua.peek l 1
						m_two <- Lua.peek l 2
						case (m_one, m_two) of
							(Just one, Just two) -> do
								params <- if top >= 3
									then do
										is3nil <- Lua.isnil l 3
										is3tbl <- Lua.istable l 3
										case (is3nil, is3tbl) of
											(True, _) -> return Nothing
											(False, True) -> do
												tbl <- toparamstable l 3
												return $ Just tbl
											(False, False) -> fail "async_submit_page parameter 3 is not a table"
									else return Nothing
								f <- async_submit_page ref one two params
								lua_log_line ref "< async_submit_page_func requested" (return ())
								push_function l (\l -> do
									(pt, puri) <- f
									lua_log_line ref ("< async_submit_page_func result " ++ (show puri)) (return ())
									Lua.pushbytestring l pt
									Lua.pushstring l (show puri)
									return 2) "async_submit_page_callback"
								return 1
							_ -> fail "Wrong parameters to async_submit_page"

			register_function "get_semirare_encounters" $ \l -> lua_log_line ref "get_semirare_encounters" $ do
				semistext <- readFile "cache/data/semirares"
				let Just semis = read_as semistext
				push_stringlist l semis
				return 1

			register_function "get_inventory_counts" $ \l -> lua_log_line ref "get_inventory_counts" $ do
				counts <- KoL.Api.getInventoryCounts ref
				push_integertable l counts
				return 1

			register_function "get_status_info" $ \l -> lua_log_line ref "get_status_info" $ do
				jsonobj <- KoL.Api.getCharStatusObj ref
				push_jsvalue l jsonobj
				return 1

			register_function "get_item_data_by_id" $ \l -> lua_log_line ref "get_item_data_by_id" $ do
				m_id <- Lua.peek l 1
				case m_id of
					Just id -> do
						(fid, _fname) <- get_item_funcs_internal ref
						case fid id of
							Just tbl -> do
								push_stringtable l tbl
								return 1
							_ -> return 0
					_ -> return 0

			register_function "get_item_data_by_name" $ \l -> lua_log_line ref "get_item_data_by_name" $ do
				m_name <- Lua.peek l 1
				case m_name of
					Just name -> do
						(_fid, fname) <- get_item_funcs_internal ref
						case fname name of
							Just tbl -> do
								push_stringtable l tbl
								return 1
							_ -> return 0
					_ -> return 0

			register_function "get_fallback_choicespoilers" $ \l -> lua_log_line ref "get_fallback_choicespoilers" $ do
				spoilertext <- readFile "cache/data/choice-spoilers"
				let Just fallback_spoilers = read_as spoilertext
				Lua.newtable l
				topidx <- Lua.gettop l
				let add_line xs = do
					case (lookup "choice number" xs, lookup "spoilers" xs) of
						(Just choicenumstr, Just spoilersstr) -> do
							case (read_as choicenumstr, read_as spoilersstr) of
								(Just choicenum, Just spoilers) -> do
									Lua.pushinteger l (fromIntegral (choicenum :: Integer))
									push_stringlist l spoilers
									Lua.settable l topidx
								_ -> throwIO $ InternalError $ "Invalid choice spoiler, id: " ++ show xs
						_ -> throwIO $ InternalError $ "Invalid choice spoiler, id: " ++ show xs
				mapM_ add_line fallback_spoilers
				return 1

			register_function "get_pulverize_groups" $ \l -> lua_log_line ref "get_pulverize_groups" $ do
				groupstext <- readFile "cache/data/pulverize-groups"
				let Just groups = read_as groupstext
				Lua.newtable l
				topidx <- Lua.gettop l
				let add_group (trgidx, (label, items)) = do
					Lua.pushinteger l trgidx
					Lua.newtable l
					gidx <- Lua.gettop l

					Lua.pushstring l "label"
					Lua.pushstring l label
					Lua.settable l gidx

					Lua.pushstring l "items"
					Lua.newtable l
					iidx <- Lua.gettop l

					mapM_ (\x -> do
						Lua.pushinteger l x
						Lua.pushboolean l True
						Lua.settable l iidx) items

					Lua.settable l gidx

					Lua.settable l topidx

				mapM_ add_group (zip [1..] groups)
				return 1

			register_function "get_faxbot_monsterlist" $ \l -> lua_log_line ref "get_faxbot_monsterlist" $ do
				mdstext <- readFile "cache/data/faxbot-monsterlist"
				let Just mds = read_as mdstext
				Lua.newtable l
				topidx <- Lua.gettop l
				let add_monsterdata (trgidx, tbl) = do
					Lua.pushinteger l trgidx
					push_stringtable l tbl
					Lua.settable l topidx

				mapM_ add_monsterdata (zip [1..] mds)
				return 1

			Lua.registerhsfunction l "get_current_kolproxy_version" get_current_kolproxy_version
			Lua.registerhsfunction l "get_latest_kolproxy_version" get_latest_kolproxy_version

			register_function "kolproxy_log_time_interval" $ \l -> do
				desc <- Lua.tostring l 1
				Lua.remove l 1
				isok <- log_time_interval ref ("lua:" ++ desc) $ Lua.pcall l 0 Lua.multret 0
-- 				isok <- Lua.pcall l 0 Lua.multret 0
				case isok of
					0 -> do
						top <- Lua.gettop l
						return $ fromIntegral top
					_ -> do
						err <- Lua.tostring l (-1)
						throwIO $ userError $ err

			Lua.registerhsfunction l "can_read_state" $ do
				x <- canReadState ref
				return $ if x then "yes" else "no"

			Lua.registerhsfunction l "parse_table_string" parse_table_string

			Lua.registerhsfunction l "get_recipes" (get_recipes ref)

			Lua.registerhsfunction l "get_raw_charpane_text" (getRawCharpaneText ref)

			let get_api_itemid_info l = lua_log_line ref "get_api_itemid_info" $ do
				m_id <- Lua.peek l 1 :: IO (Maybe Int)
				case m_id of
					Just itemid -> do
-- 						jsonobj <- KoL.Api.getItemInfoObj itemid ref
-- 						push_jsvalue l jsonobj
-- 						return 1
						f <- KoL.Api.asyncGetItemInfoObj itemid ref
						push_function l (\l -> do
							jsonobj <- f
							push_jsvalue l jsonobj
							return 1) "get_api_itemid_info_callback"
						return 1
					_ -> return 0

			case level of
				"whenever" -> do
					Lua.registerhsfunction l "get_character_name" (get_ref_playername ref)
					Lua.registerhsfunction l "get_player_id" (get_player_id ref)
					Lua.registerhsfunction l "set_chat_state" (set_state_whenever ref)
					Lua.registerhsfunction l "get_chat_state" (get_state_whenever ref)
				"process" -> do
					Lua.registerhsfunction l "set_state" (set_state ref)
					Lua.registerhsfunction l "get_state" (get_state ref)
					Lua.registerhsfunction l "reset_fight_state" (uglyhack_resetFightState ref)
					register_function "get_api_itemid_info" get_api_itemid_info
				"printer" -> do
					Lua.registerhsfunction l "get_state" (get_state ref)
					Lua.registerhsfunction l "reset_fight_state" (uglyhack_resetFightState ref)
					register_function "get_api_itemid_info" get_api_itemid_info
				"automate" -> do
					Lua.registerhsfunction l "set_state" (set_state ref)
					Lua.registerhsfunction l "get_state" (get_state ref)
					register_function "submit_page" submit_page_func
					register_function "async_submit_page" async_submit_page_func
					register_function "get_api_itemid_info" get_api_itemid_info
				"intercept" -> do
					Lua.registerhsfunction l "set_state" (set_state ref)
					Lua.registerhsfunction l "get_state" (get_state ref)
					register_function "submit_page" submit_page_func
					register_function "async_submit_page" async_submit_page_func
					register_function "get_api_itemid_info" get_api_itemid_info
				_ -> throwIO $ InternalError $ "Invalid Lua run level: " ++ level

		Lua.registerhsfunction l "kolproxy_debug_print" (\x -> lua_log_line ref ("kolproxy_debug_print: " ++ x) (return ()))

		Lua.safeloadstring l "local dtb = debug.traceback; return function(e) kolproxy_debug_traceback = dtb(\"\", 2) return e end"
		Lua.pcall l 0 1 0 -- put error handling function on stack=1

		c <- Lua.safeloadstring l code
		return (l, c)

-- 	putStrLn $ "> run_lua_code/go: " ++ (filename) ++ " [" ++ (show $ lookup "path" vars) ++ "]"
	ret <- case c of
		0 -> do
-- 			sd "running"
			moo <- log_time_interval ref ("do lua loading code: " ++ filename) $ Lua.pcall l 0 Lua.multret 1 -- returns on stack=2+
			case moo of
				0 -> do
					Lua.setglobal l "kolproxy_stored_wrapped_function"
					return $ Right l
				_ -> do
					putStrLn $ "lua error!"
					top <- Lua.gettop l
					putStrLn $ "lua error, top = " ++ (show top)
					err <- Lua.tostring l (-1)
					putStrLn $ "lua error: " ++ err
					Lua.getglobal l "kolproxy_debug_traceback" -- load traceback on stack=3
					traceback <- Lua.tostring l 3
					putStrLn $ "traceback: " ++ traceback
					return $ Left $ ("error running loading code (" ++ filename ++ "):\n" ++ err, traceback)
		errnum -> do
			putStrLn $ "luaload error!"
			top <- Lua.gettop l
			putStrLn $ "luaload error, top = " ++ (show top)
			err <- Lua.tostring l (-1)
			putStrLn $ "luaload error: " ++ err
			return $ Left $ ("error " ++ (show errnum) ++ " loading code (" ++ filename ++ "):\n" ++ err, "")
-- 	putStrLn $ "< run_lua_code/go: " ++ (filename) ++ " [" ++ (show $ lookup "path" vars) ++ "]"
-- 	log_time_interval ref "close lua" $ Lua.close l
-- 	sd "done"
-- 	putStrLn $ "< run_lua_code: " ++ (filename) ++ " [" ++ (show $ lookup "path" vars) ++ "]"
	return ret

get_cached_lua_instance_for_code level filename ref codebit = do
	-- TODO! CLOSE LUA INSTANCES!!
	let make_lsetup = log_time_interval ref ("setup lua instance: " ++ filename ++ "|" ++ level) $ setup_lua_instance level filename ref

	canread <- canReadState ref
	let use_cache = case (canread, level) of
		(_, "whenever") -> False
		(False, _) -> False
		_ -> True

	if use_cache
		then do
			-- TODO: Wrap in MVar instead of IORef?
			insts <- readIORef (luaInstances $ sessionData_ $ otherstuff_ $ ref)
			case Data.Map.lookup (filename, level) insts of
				-- Use MVar to only run one piece of code in an instance at a time, e.g. sequence one processor after another serially.
				Just old -> withMVar old (\x -> codebit (Right x))
				_ -> do
					either_l_setup <- make_lsetup
					case either_l_setup of
						Right l_setup -> do
							mv_l <- newMVar l_setup
							let newinsts = Data.Map.insert (filename, level) mv_l insts
							writeIORef (luaInstances $ sessionData_ $ otherstuff_ $ ref) newinsts
							withMVar mv_l (\x -> codebit (Right x))
						Left err -> return $ Left err
		else do
-- 			putStrLn $ "DEBUG: not cached " ++ show (canread, level) ++ " | " ++ show filename
			either_l_setup <- make_lsetup
			case either_l_setup of
				Right l_setup -> do
					x <- codebit (Right l_setup)
					log_time_interval ref "close lua" $ Lua.close l_setup
					return x
				Left err -> return $ Left err

run_lua_code level filename ref vars pagetext = do
	get_cached_lua_instance_for_code level filename ref $ \l_setup -> do
		case lookup "text" vars of
			Just text -> (lua_log_line ref ("run_lua_code:" ++ filename ++ " [" ++ (show $ lookup "path" vars) ++ ", param length " ++ (show $ length $ fromJust $ lookup "raw_input_params" vars) ++ ", text length " ++ (show $ length text) ++ "]") (return ())) `catch` (\e -> putStrLn $ "run_lua_code error: " ++ (show (e :: SomeException)))
			_ -> (lua_log_line ref ("run_lua_code:" ++ filename ++ " [" ++ (show $ lookup "path" vars) ++ ", param length " ++ (show $ length $ fromJust $ lookup "raw_input_params" vars) ++ ", no text]") (return ())) `catch` (\e -> putStrLn $ "run_lua_code error: " ++ (show (e :: SomeException)))
	-- 	putStrLn $ "> run_lua_code: " ++ (filename) ++ " [" ++ (show $ lookup "path" vars) ++ "]"
	-- 	let sd x = segfault_debug $ printf "lua:%s[%s] %s | %s" x level filename (show [(x, take 50 y) | (x, y) <- vars])
	-- 	sd "loading"

		ret <- case l_setup of
			Right l -> do
				log_time_interval ref "prepare for lua code" $ do
					top <- Lua.gettop l
					when (top > 1) $ do
						putStrLn $ "Warning: Lua top is " ++ show top
						Lua.pop l (top - 1)
					Lua.getglobal l "kolproxy_stored_wrapped_function"
					push_stringtable l vars
					Lua.pushstring l "text"
					Lua.pushbytestring l pagetext
					Lua.settable l (-3)
				moo_two <- log_time_interval ref ("run lua code: " ++ filename) $ Lua.pcall l 1 Lua.multret 1 -- returns on stack=2+
				log_time_interval ref "retrieving results" $ case moo_two of
					0 -> do
						top <- Lua.gettop l
						rets <- mapM (\x -> do
							isstring <- Lua.isstring l x
							if isstring
								then do
									r <- Lua.tobytestring l x
-- 									putStrLn $ "r#" ++ (show x) ++ ": " ++ (show r)
									return $ Just r
								else return Nothing) [2..top]
						Lua.pop l (top - 1)
						return $ Right rets
					_ -> do
						putStrLn $ "lua-call error!"
						top <- Lua.gettop l
						putStrLn $ "lua-call error, top = " ++ (show top)
						err <- Lua.tostring l (-1)
						putStrLn $ "lua-call error: " ++ err
						Lua.getglobal l "kolproxy_debug_traceback" -- load traceback on stack=3
						traceback <- Lua.tostring l 3
						putStrLn $ "call-traceback: " ++ traceback
						Lua.pop l (top - 1)
						return $ Left $ ("error running code (" ++ filename ++ "):\n" ++ err, traceback)
			Left err -> do
				return $ Left err
		return ret

-- make_test_filename filename inuri ineffuri = do
-- 	let accepted x
-- 		| isAsciiUpper x = True
-- 		| isAsciiLower x = True
-- 		| isDigit x = True
-- 		| (x == '.') = True
-- 		| otherwise = False
-- 	let uri = filter accepted (uriPath inuri)
-- 	let effuri = filter accepted (uriPath ineffuri)
-- 	return (uri ++ "-" ++ effuri ++ "-" ++ (takeBaseName filename))

-- runScript_generate_test filename ref uri effuri pagetext allparams = do
-- 	let stateref = state ref
-- 	(_, prestate) <- readIORef stateref

-- 	code <- readFile filename
-- 	ret <- run_lua filename code ref uri effuri pagetext allparams "(error running code)"

-- 	(_, poststate) <- readIORef stateref

-- 	if or [ret /= Right pagetext, prestate /= poststate]
-- 		then do
-- 			fn <- make_test_filename filename uri effuri
-- 			(fp, h) <- openTempFile "test/" ("test-" ++ fn ++ "-.data") -- TODO: OPENFILE
-- 			let writedata = [
-- 				("filename", show filename),
-- 				("uri", show uri),
-- 				("effuri", show effuri),
-- 				("prestate", show prestate),
-- 				("poststate", show poststate),
-- 				("pagetext", show pagetext),
-- 				("output", show ret)]
-- 			hPutStr h (show writedata)
-- 			hClose h
-- 		else return ()
-- 	return ret

runProcessScript ref uri effuri pagetext allparams = do
	let vars = [("path", uriPath effuri), ("query", uriQuery effuri), ("raw_input_params", show (allparams :: [(String, String)])), ("requestpath", uriPath uri), ("requestquery", uriQuery uri)]
	rets <- run_lua_code "process" "scripts/process-page.lua" ref vars pagetext
	return $ case rets of
		Right [Just t] -> Right $ t
		Right xs -> Left ("Lua process call error, return values = " ++ (show xs), "")
		Left err -> Left err

runPrinterScript ref uri effuri pagetext allparams = do
	let vars = [("path", uriPath effuri), ("query", uriQuery effuri), ("raw_input_params", show (allparams :: [(String, String)])), ("requestpath", uriPath uri), ("requestquery", uriQuery uri)]
	rets <- run_lua_code "printer" "scripts/printer.lua" ref vars pagetext
	return $ case rets of
		Right [Just t] -> Right $ t
		Right xs -> Left ("Lua printer call error, return values = " ++ (show xs), "")
		Left err -> Left err

runChatScript ref uri effuri pagetext allparams = do
	let vars = [("path", uriPath effuri), ("query", uriQuery effuri), ("raw_input_params", show (allparams :: [(String, String)])), ("requestpath", uriPath uri), ("requestquery", uriQuery uri)]
	rets <- run_lua_code "whenever" "scripts/chat.lua" ref vars pagetext
	return $ case rets of
		Right [Just t] -> Right $ t
		Right xs -> Left ("Lua chat call error, return values = " ++ (show xs), "")
		Left err -> Left err

runAutomateScript ref uri effuri pagetext allparams = do
	let vars = [("path", uriPath effuri), ("query", uriQuery effuri), ("raw_input_params", show allparams), ("requestpath", uriPath uri), ("requestquery", uriQuery uri)]
	rets <- run_lua_code "automate" "scripts/automate.lua" ref vars pagetext
	return $ case rets of
		Right [Just t] -> Right t
		Right xs -> Left ("Lua automate call error, return values = " ++ (show xs), "")
		Left err -> Left err

runInterceptScript ref uri allparams reqtype = do
	let vars = [("requestpath", uriPath uri), ("requestquery", uriQuery uri), ("raw_input_params", show allparams), ("request_type", reqtype)]
	rets <- run_lua_code "intercept" "scripts/intercept.lua" ref vars (Data.ByteString.Char8.pack "")
	return $ case rets of
		Right [Just t, Just u] -> Right (t, mkuri $ Data.ByteString.Char8.unpack $ u)
		Right xs -> Left ("Lua intercept call error, return values = " ++ (show xs), "")
		Left err -> Left err
